
# THIS FILE IS GENERATED FROM THEANO SETUP.PY
short_version = '0.4.1'
version = '0.4.1'
hg_revision = '65d0434e835d+'
full_version = '0.4.1.dev-%(hg_revision)s' % {'hg_revision': hg_revision}
release = True

if not release:
    version = full_version
